import json
import random
import sys

import vlns
import solver

from clingo.symbol import Function, Number
from collections import namedtuple
import argparse
import logging
logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, format='%(asctime)-15s [%(process)d:%(processName)s] %(module)s:%(lineno)d %(levelname)s: %(message)s')

Instance = namedtuple('Instance', ['n_machines','n_jobs', 'duration', 'release', 'capable', 'setup'])
Solution = namedtuple('Solution', ['assigned','next', 'cost'])

program = '''
1 { assigned(J,M) : capable(M,J) } 1 :- job(J).
1 { first(J,M) : capable(M,J) } 1 :- assigned(_,M).
1 { last(J,M) : capable(M,J) } 1 :- assigned(_,M).
:- first(J,M), not assigned(J,M).
:- last(J,M), not assigned(J,M).
1 { next(J1,J2,M) : capable(M,J1), J1 != J2  } 1 :- assigned(J2,M), not first(J2,M).
1 { next(J1,J2,M) : capable(M,J2), J1 != J2  } 1 :- assigned(J1,M), not last(J1,M).
:- next(J1,J2,M), not assigned(J1,M).
:- next(J1,J2,M), not assigned(J2,M).
reach(J1,M) :- first(J1,M).
reach(J2,M) :- reach(J1,M), next(J1,J2,M).
:- assigned(J1,M), not reach(J1,M).
&diff{ 0 - compl(J2) } <= -(T+D+S) :- next(J1,J2,M), setup(J1,J2,M,S), duration(J2,M,D), release(J2,M,T).
&diff{ compl(J1) - compl(J2) } <= -(P+S) :- next(J1,J2,M), setup(J1,J2,M,S), duration(J2,M,P).
&diff{ compl(J1) - makespan } <= 0 :- job(J1).
&diff{ compl(J1) - compl(J2) } <= -P :- next(J1,J2,M), duration(J2,M,P).
&diff{ 0 - compl(J1) } <= -(T+D) :- assigned(J1,M), duration(J1,M,D), release(J1,M,T).

#show next/3.
#show assigned/2.
'''

class PmspVLNS(vlns.AbstractClingoVLNS):

    def __init__(self, move_timeout=15, relax_factor=0.2, use_greedy=True, forget_on_shot=False, seed=None):
        super().__init__(move_timeout=move_timeout, strict_bound_prob=1.0, use_greedy=use_greedy)
        self.__relax_factor = relax_factor
        self.__forget_on_shot = forget_on_shot
        self.__seed = seed
        self.__leading_machine = None

    def _program(self):
        return program

    def __build_fact(self, name, *args):
        return name + '(' + ','.join([ str(a) for a in args ]) + ').'

    def _instance_to_facts(self, instance):
        facts = []
        for j in range(instance.n_jobs):
            facts += [ self.__build_fact('job', j) ]

        for m in range(instance.n_machines):
            facts += [ self.__build_fact('machine', m) ]

        for j in range(instance.n_jobs):
            for k in range(instance.n_jobs):
                for m in range(instance.n_machines):
                    facts += [ self.__build_fact('setup', j, k, m, instance.setup[(j,k,m)]) ]

        for j in range(instance.n_jobs):
            for m in range(instance.n_machines):
                facts += [ self.__build_fact('duration', j, m, instance.duration[(j,m)]) ]
                facts += [ self.__build_fact('release', j, m, instance.release[(j,m)]) ]
                if m in instance.capable[j]:
                    facts += [ self.__build_fact('capable', m, j) ]

        return ''.join(facts)

    def _construct_initial_solution(self, instance):
        jobs = list(range(instance.n_jobs))
        jobs.sort(
            key=lambda j: max(max(0, instance.release[(j,m)]) for m in range(instance.n_machines) if m in instance.capable[j])
        )

        current_span = { m: 0 for m in range(instance.n_machines) }

        assigned = {}
        next = {}
        current_last = {}

        for j in jobs:
            assigned_to = None
            best_span = None
            for m in instance.capable[j]:
                new_span = max(current_span[m], instance.release[(j,m)]) + instance.duration[(j,m)]
                if m in current_last:
                    new_span += instance.setup[(current_last[m], j, m)]

                if assigned_to == None or new_span < best_span:
                    assigned_to = m
                    best_span = new_span

            if assigned_to in current_last:
                next[current_last[assigned_to]] = j
            assigned[j] = assigned_to
            current_last[assigned_to] = j
            current_span[assigned_to] = best_span

        cost = self.__calculate_cost(assigned, next, instance)

        if not cost:
            raise Exception('Greedy returned an invalid solution!')

        return Solution(assigned, next, cost)


    def _solution_to_assumptions(self, solution):
        assumptions = []
        for j,m in solution.assigned.items():
            assumptions += [ Function('assigned', [Number(j), Number(m)]) ]

        for j,k in solution.next.items():
            assumptions += [ Function('next', [Number(j), Number(k), Number(solution.assigned[j])]) ]

        return assumptions

    def _get_move_assumptions(self, instance, incumbent):
        assumptions = []
        selection_size = round(instance.n_jobs * self.__relax_factor)

        if random.uniform(0.0,1.0) < 0.5:
            logger.debug('using job neighborhood')
            selected_jobs = random.sample([ j for j in range(instance.n_jobs) ], selection_size)
        else:
            logger.debug('using machine neighborhood')
            selected_jobs = [ j for j in range(instance.n_jobs) if incumbent.assigned[j] == self.__leading_machine ]
            other_machine = random.randint(0, instance.n_machines - 1)
            if other_machine >= self.__leading_machine:
                other_machine += 1
            selected_jobs += [ j for j in range(instance.n_jobs) if incumbent.assigned[j] == other_machine ]
            if len(selected_jobs) > selection_size:
                selected_jobs = random.sample(selected_jobs, selection_size)

        machines = { incumbent.assigned[j] for j in selected_jobs }
        if not self.__leading_machine in machines:
            # leading machine not in selection
            # remove one job
            selected_jobs = selected_jobs[:-1]
            # add random job from leading machine
            selected_jobs += random.sample([ j for j in range(instance.n_jobs) if incumbent.assigned[j] == self.__leading_machine ], 1)

        for j,m in incumbent.assigned.items():
            if not j in selected_jobs:
                assumptions += [ Function('assigned', [Number(j), Number(m)]) ]

        for j,k in incumbent.next.items():
            if not (j in selected_jobs or k in selected_jobs):
                assumptions += [ Function('next', [Number(j), Number(k), Number(incumbent.assigned[j])]) ]

        return assumptions

    def _model_to_solution(self, model, instance):
        assigned = {}
        next = {}
        for s in model.shown:
            if s.match('assigned', 2):
                assigned[s.arguments[0].number] = s.arguments[1].number
            elif s.match('next', 3):
                next[s.arguments[0].number] = s.arguments[1].number

        # calculate makespan and check solution
        cost = self.__calculate_cost(assigned, next, instance)

        if not cost:
            raise Exception('VLNS returned an invalid solution!')

        solution = Solution(assigned, next, cost)

        return solution


    def __calculate_cost(self, assigned, next, instance):
        prev = { k: j for j,k in next.items() }

        completion_time = {}
        makespan = None
        for first in [ k for k in range(instance.n_jobs) if not k in prev ]:
            j = first
            prev_finished = 0
            while j != None:
                if not j in assigned:
                    return

                start_time = max(prev_finished, instance.release[(j,assigned[j])])
                setup = 0
                if j in prev:
                    setup = instance.setup[(prev[j],j,assigned[j])]

                    if not prev[j] in assigned or assigned[j] != assigned[prev[j]]:
                        return

                end_time = start_time + setup + instance.duration[(j,assigned[j])]
                completion_time[j] = end_time
                if makespan == None or makespan[1] < completion_time[j]:
                    makespan = (assigned[j], completion_time[j])

                prev_finished = end_time
                if j in next:
                    j = next[j]
                else:
                    j = None

        for j in range(instance.n_jobs):
            if not j in completion_time:
                return

        self.__leading_machine = makespan[0]
        return makespan[1]


    def _internal_solver(self):
        return solver.ClingoDl(minimize_variable='makespan', forget_on_shot=self.__forget_on_shot, seed=self.__seed)


def main(*args):

    parser = argparse.ArgumentParser(
            description='A solver for PMSP with release dates.')
    parser.add_argument('-i', '--input', type=str, metavar='INSTANCE', help='the input instance', required=True)
    parser.add_argument('-t', '--timeout', type=int, metavar='TIMEOUT', help='the timeout in seconds for the search', required=True)
    parser.add_argument('-r', '--relax-factor', type=float, metavar='FACTOR', help='the percentage of jobs relxed at each lns iteration (default: %(default)s)', default=0.2)
    parser.add_argument('-m', '--move-timeout', type=float, metavar='MOVE_TIMEOUT', help='the timeout in seconds for each lns iteration (default: %(default)s, -1 means no timeout)', default=15)
    parser.add_argument('-s', '--seed', type=int, metavar='SEED', help='the seed used for the random generator', default=None)
    parser.add_argument('-g', '--use-greedy', action='store_true', help='whether or not the greedy heuristic should be used to obtain an initial solution')
    parser.set_defaults(use_greedy=False)
    parser.add_argument('-f', '--forget-on-shot', action='store_true', help='whether or not the previous state should be forgotten on each new lns iteration')
    parser.set_defaults(forget_on_shot=False)

    args = parser.parse_args()

    relax_factor = args.relax_factor
    if not (relax_factor > 0 and relax_factor <= 1):
        raise ValueError('invalid relax factor')

    timeout = args.timeout
    if timeout <= 0:
        raise ValueError('invalid timeout')

    move_timeout = args.move_timeout
    if timeout <= -1:
        raise ValueError('invalid move timeout')

    if args.seed == None:
        seed_value = random.randrange(sys.maxsize)
    else:
        seed_value = args.seed

    logger.info('Seed value: %i' % seed_value)

    random.seed(seed_value)

    solver = PmspVLNS(move_timeout=move_timeout, relax_factor=relax_factor, use_greedy=args.use_greedy, forget_on_shot=args.forget_on_shot, seed=seed_value)
    instance = load_json_instance(args.input)
    timelimit = timeout



    solution = solver.solve(instance, timelimit)

    print(solution)


def load_json_instance(instance_path):
    with open(instance_path, 'r') as file:
        json_object = json.load(file)
        n_machines = json_object['m']
        n_jobs = json_object['n']
        duration = { (j,m) : json_object['duration'][j][m] for j in range(n_jobs) for m in range(n_machines) }
        release = { (j,m) : json_object['release'][j][m] for j in range(n_jobs) for m in range(n_machines) }
        capable = { j : set(json_object['capable'][j]) for j in range(n_jobs) }
        setup = { (j,k,m) : json_object['setup'][j][k][m] for m in range(n_machines) for j in range(n_jobs) for k in range(n_jobs) }

    return Instance(n_machines, n_jobs, duration, release, capable, setup)


if __name__ == '__main__':
    args = sys.argv
    main(*args)
